/*
 * file StpProperty.java
 *
 * Licensed Materials - StpProperty of IBM
 * Restricted Materials of IBM 
 *
 * com.ibm.rational.wvcm.stp.StpProperty
 *
 *  Copyright IBM Corporation 2004, 2008.  All Rights Reserved.
 * Note to U.S. Government Users Restricted Rights:  Use, duplication or 
 * disclosure restricted by GSA ADP  Schedule Contract with IBM Corp.
 */

package com.ibm.rational.wvcm.stp;

import static com.ibm.rational.wvcm.stpex.StpExBase.METADATA_NAMESPACE;

import java.util.ArrayList;
import java.util.List;

import javax.wvcm.PropertyNameList;
import javax.wvcm.Resource;
import javax.wvcm.WvcmException;
import javax.wvcm.PropertyNameList.PropertyName;

import com.ibm.rational.wvcm.stp.cq.CqFieldValue.ValueType;
import com.ibm.rational.wvcm.stpex.StpExBase;
import com.ibm.rational.wvcm.stpex.StpExEnumeration;
import com.ibm.rational.wvcm.stpex.StpExEnumerationBase;

/**
 * A property is a named collection of metadata associated with a resource. The
 * collection is named by its PropertyName.
 * 
 * <h2>Meta-Properties</h2>
 * 
 * <p>
 * Each metadata component is called a meta-property and is a (name, value)
 * pair. That is, a property is a named collection of meta-properties. The name
 * of a meta-property is denoted by a {@link MetaPropertyName MetaPropertyName}
 * object. The value is an arbitrary Object.
 * </p>
 * 
 * <p>
 * The term "meta-property" is used here rather loosely. Meta-properties specify
 * more than just static attributes of a property (such as its type or name),
 * but often define very dynamic aspects of the property (e.g., what resource it
 * came from) or its value (e.g. whether or not the value is null). There is
 * even a VALUE meta-property, whose value is the value of the property.
 * </p>
 * 
 * <p>
 * Just like a PropertyName, a MetaPropertyName is composed from a namespace and
 * a simple name. (In fact, to simplify this API, MetaPropertyName is a subclass
 * of javax.wvcm.PropertyNameList.PropertyName.) The PropertyName and
 * MetaPropertyName namespaces cannot overlap.
 * </p>
 * 
 * <p>
 * Every property has the meta-properties defined in this class; namely
 * {@link #NAME NAME}, {@link #NAMESPACE NAMESPACE}, {@link #PROPERTY_NAME
 * PROPERTY_NAME}, {@link #RESOURCE RESOURCE} and {@link #VALUE VALUE}. Each
 * property may have additional meta-properties of its own as well.
 * </p>
 * 
 * <p>
 * When we speak of the "<i>value of a property</i>", we are actually
 * referring to the value of the VALUE meta-property of the property.
 * </p>
 * 
 * <h2>Requesting Meta-properties</h2>
 * 
 * <p>
 * An StpProperty object is a client-side proxy for a property; just as a
 * Resource object is a client-side proxy for a resource. Like a resource proxy,
 * each StpProperty instance defines only those meta-properties that have been
 * explicitly requested from the server. (The NAME, NAMESPACE, and PROPERTY_NAME
 * meta-properties are an exception since they are required to construct a
 * StpProperty instance and are always present.)
 * </p>
 * 
 * <p>
 * An StpProperty instance does not provide an interface for changing the value
 * of a meta-property. Clients can change only property values, which may have
 * the side-effect of changing other meta-properties for that property or
 * meta-properties for other properties.
 * </p>
 * 
 * <p>
 * Property meta-properties are available only in the context of a specific
 * resource and the request for property meta-properties is tied directly to
 * that resource using the same mechanism used to request properties from that
 * resource.
 * </p>
 * 
 * <h3>MetaPropertyName Objects in a NestedPropertyName</h3>
 * 
 * <p>
 * To request a meta-property of a property, a MetaPropertyName object for the
 * meta-property is included in the PropertyRequest argument to
 * PropertyName.nest(). MetaPropertyName objects and PropertyName objects can
 * coexist in the same nested PropertyRequest; the nested MetaPropertyName
 * objects request meta-properties of the property, while the nested ProperyName
 * objects request properties of the resource named by the value of the
 * property.
 * </p>
 * 
 * <p>
 * To request the meta-properties of a property-valued property, the
 * MetaPropertyNames must be nested within a NestedPropertyName. The root
 * meta-property of this request will be StpProperty.VALUE and the desired
 * meta-properties of the value will be nested in the PropertyRequest of the
 * NestedPropertyName.
 * </p>
 * 
 * <p>
 * Examples
 * </p>
 * 
 * <pre>
 *  PropertyRequest wanted=new PropertyRequest(
 *    // The value of the DISPLAY_NAME property
 *    Resource.DISPLAY_NAME,
 *    Resource.WORKSPACE_FOLDER_LIST.nest(
 *        // The type of the WORKSPACE_FOLDER_LIST property
 *        StpProperty.TYPE,
 *        // The value of the CHILD_LIST property for each folder
 *        // named by the WORKSPACE_FOLDER_LIST
 *        Folder.CHILD_LIST),  
 *    Resource.INVALID_PROPERTIES.nest(
 *           // The type of the INVALID_PROPERTIES property
 *           StpProperty.TYPE,
 *           StpProperty.VALUE.nest(
 *               // The type of each property named by the value of the 
 *               // INVALID_PROPERTIES property.
 *               StpProperty.TYPE,
 *               // The value of each property named by the value of the 
 *               // INVALID_PROPERTIES property.
 *               StpProperty.VALUE),
 *  );
 * </pre>
 * 
 * <h3>The VALUE Meta-Property</h3>
 * 
 * <p>
 * Every property has a VALUE meta-property. The use of a PropertyName by itself
 * implicitly requests this meta-property. A NestedPropertyName containing
 * nested ProperyNames also implicitly requests the VALUE meta-property (since
 * it would be impossible to request properties of the value without also
 * requesting the value itself).
 * </p>
 * 
 * <p>
 * On the other hand, a NestedPropertyName that contains only nested
 * MetaPropertyNames must include StpProperty.VALUE if the value of the property
 * is desired. This makes it possible for a client to request information about
 * a property (such as is size) without actually retrieving the value itself.
 * </p>
 * 
 * <h2>Accessing Meta-Properties</h2>
 * 
 * <p>
 * Like property values, meta-properties retrieved from the server are stored in
 * the resource proxy and accessed via methods defined in the StpResource class.
 * The principal method is {@link
 * com.ibm.rational.wvcm.stp.StpResource#getMetaProperties(PropertyNameList.PropertyName)
 * Resource.getMetaProperties(PropertyName)}, which returns an StpProperty object
 * containing all of the meta-properties for a given property retrieved from the
 * server and diagnostic information for those meta-properties that couldn't be
 * retrieved.
 * </p>
 * 
 * <p>
 * As long as the PropertyName was mentioned in the request and that property is
 * defined by the resource, getMetaProperties will return a StpProperty object
 * for that PropertyName. An exception is thrown only if the property is not
 * defined by the resource or no meta-property of the property was requested.
 * </p>
 * 
 * <p>
 * {@link javax.wvcm.Resource#getProperty(PropertyNameList.PropertyName) 
 * Resource.getProperty(PropertyName)} returns the value of the VALUE
 * meta-property of the property and will throw an exception if that
 * meta-property was not requested or the server reported an error in attempting
 * to retrieve the value. If the value of a property is, itself, a property,
 * getProperty will return an StpProperty object.
 * </p>
 * 
 * <p>
 * From the StpProperty object, the client can access the retrieved
 * meta-properties, either generically using the {@link
 * StpProperty#getMetaProperty(StpProperty.MetaPropertyName) getMetaProperty()}
 * method and an explicit MetaPropertyName or using the specialized methods,
 * such as {@link #getName getName} and {@link #getValue getValue}, where the
 * MetaPropertyName is implicit. Subclasses of Resource may define subclasses of
 * StpProperty and provide typed accessors for them. The subclasses of
 * StpProperty may, in turn, define more typed accessors for meta-properties
 * they know about. (See, for example
 * {@link com.ibm.rational.wvcm.stp.cq.CqRecord Record} and
 * {@link com.ibm.rational.wvcm.stp.cq.CqFieldValue CqFieldValue}.)
 * </p>
 * <p>
 * The type parameter of an StpProperty instance specifies the type of the 
 * property's VALUE meta-property.
 */
public interface StpProperty<T>
{
    /**
     * A List of Properties.
     */
    public static class List<S extends StpProperty>
        extends ArrayList<S>
    {
        /**
         * 
         */
        private static final long serialVersionUID = 1181176128014802915L;

        /**
         * Returns the names of the properties in the list
         * 
         * @return A PropertyNameList containing a PropertyName for each element
         *         of the list.
         */
        public PropertyNameList getPropertyNameList()
        {
            int            n = size();
            PropertyName[] names = new PropertyName[n];

            for (int i = 0; i < n; ++i)
                names[i] = get(i).getPropertyName();

            return new PropertyNameList(names);
        }

        /* (non-Javadoc)
         * @see java.util.ArrayList#add(java.lang.Object)
         */
        @Override
        public boolean add(S arg0)
        {
            return super.add(arg0);
        }
    }

    /**
     * An object representing the name of a meta-property. This class extends
     * PropertyName so that MetaPropertyName objects may be included in a
     * PropertyRequest. Otherwise, MetaPropertyName and PropertyName are not
     * interchangeable. The type parameter specifies the type of meta-property's
     * value.
     */
    public static class MetaPropertyName<V>
        extends PropertyName<V>
    {
        /**
         * Creates a new MetaPropertyName from a simple name and a specified
         * namespace.
         *
         * @param  namespace  the namespace for this MetaPropertyName. May not
         *                    be <b>null</b>.
         * @param  name  the name for this property. May not be <b>null</b>.
         */
        public MetaPropertyName(
            String namespace,
            String name)
        {
            super(namespace, name);
        }
        
        /**
         * Parameterized method for constructing MetaPropertyName<?>
         * 
         * @param <U> The value type for the MetaPropertyName
         * @param namespace The namespace for this MetaPropertyName. May not be
         *            <b>null</b>.
         * @param name the name for this property. May not be <b>null</b>.
         * @return A new instanceof MetaPropertyName using the given namespace
         *         and name
         */
        public static final <U> MetaPropertyName<U> build(String namespace,
                                                          String name)
        {
            return new MetaPropertyName<U>(namespace, name);
        }
    }

    /**
     * Returns the value of the given meta-property defined by this StpProperty
     * instance.
     * 
     * @param name A MetaPropertyName object identifying the meta-property
     *            desired.
     * 
     * @return An Object representing the value of the meta-property requested.
     * 
     * @throws WvcmException if this StpProperty instance does not define a
     *             value for the named meta-property.
     */
    public <U> U getMetaProperty(MetaPropertyName<U> name)
        throws WvcmException;

    /**
     * Returns the names of meta-properties available on this client StpProperty
     * proxy. This is the analog of Resource.propertyNameList. A meta-property
     * is available if it would not throw an exception if retrieved from this
     * StpProperty. Failing to retrieve a meta-property would cause a
     * meta-property name to not appear in the metaPropertyNames result and
     * forgetting or removing a meta-property would remove that property from
     * the metaPropertyNames result.
     * 
     * @return the list of MetaPropertyName objects identifying the
     *         meta-properties available on this client StpProperty proxy.
     * 
     * @see javax.wvcm.Resource#propertyNameList()
     */
    public MetaPropertyName<?>[] metaPropertyNames();

    /**
     * Returns an array of StpPropertyException objects, each of which
     * represents a meta-property whose value was requested but could not be
     * successfully retrieved.
     * 
     * @return An array containing an StpPropertyException object for each
     *         meta-property that could not be retrieved. The PropertyName field
     *         of the PropertyException identifies the meta-property that could
     *         not be retrieved.
     */
    public StpPropertyException[] metaPropertyExceptions();
    
    /**
     * Returns the StpPropertyException that would be thrown if getMetaProperty
     * were invoked on a specified meta-property.
     * 
     * @param name The name of the meta-property whose retrieval status is being
     *            interrogated.
     * @return an StpPropertyException is returned if the named meta-property
     *         could not be retrieved or if it's retrieval was not requested;
     *         otherwise <b>null</b>.
     */
    public StpPropertyException metaPropertyException(MetaPropertyName<?> name);
    
    /**
     * A list of all the meta-property names defined for this property. This is
     * the analog of Resource.doGetPropertyNameList. Since properties may not be
     * addressed directly, this analogous capability is expressed as a
     * meta-property.
     */
    public static final MetaPropertyName<PropertyNameList> META_PROPERTY_NAMES =
        new MetaPropertyName<PropertyNameList>(METADATA_NAMESPACE,
                                               "meta-property-names");

    /**
     * Returns the value of the {@link #META_PROPERTY_NAMES META_PROPERTY_NAMES}
     * meta-property as defined by this StpProperty instance.
     *
     * @return  A PropertyRequest containing a MetaPropertyName for each meta-
     *          property defined by this StpProperty.
     *
     * @throws  WvcmException  if this StpProperty instance does not define a value
     *                         for the {@link
     *                         #META_PROPERTY_NAMES META_PROPERTY_NAMES}
     *                         meta-property.
     */
    public PropertyNameList getMetaPropertyNames()
        throws WvcmException;

    /**
     * The meta-property whose value is the value of the property. (The type of
     * the value is specified by the TYPE meta-property.) This meta-property is
     * defined for all properties.
     */
    public static final MetaPropertyName<?> VALUE =
        new MetaPropertyName(StpExBase.METADATA_NAMESPACE, "value");

    /**
     * Returns the value of the {@link #VALUE VALUE} meta-property as defined by
     * this StpProperty instance.
     * 
     * @return An Object representing the value of the property represented by
     *         this StpProperty instance.
     * 
     * @throws WvcmException if this StpProperty instance does not define a
     *             value for the {@link #VALUE VALUE} meta-property.
     */
    public T getValue()
        throws WvcmException;
    
    /**
     * An enumeration of the possible property types. Subclasses may extend this
     * list to include any new data types introduced by that subclass.
     */
    public enum Type
        implements StpExEnumeration
    {
        /** An initial value for Type variables */
        UNKNOWN,

        /** An integer in the range -9999999999 to 9999999999 */
        INTEGER,

        /**
         * A GMT date/time between 12:00:01 AM January 1, 1970 and 11:59:59 PM
         * January 18, 2038.
         */
        DATE_TIME,

        /** A proxy for a Resource of a specific ResourceType. */
        RESOURCE,

        /** A list of proxies for a Resource of a specific ResourceType */
        RESOURCE_LIST,

        /** A floating point value */
        FLOAT,

        /** A generic string value */
        STRING,

        /** An StpLocation property value. */
        LOCATION,

        /** A Locale property value. */
        LOCALE,

        /** A property value that is a list of Strings. */
        STRING_LIST,

        /** A long property value. */
        LONG,

        /** An double property value. */
        DOUBLE,

        /** A boolean property value. */
        BOOL,

        /** A ResourceType property value */
        RESOURCE_TYPE,

        /** A property value derived from StpExEnumerationBase */
        ENUMERATED_TYPE,

        /** A list of values derived from StpExEnumerationBase */
        ENUMERATED_LIST,

        /** A PropertyName property value */
        PROPERTY_NAME,

        /** A PropertyNameList property value */
        PROPERTY_NAME_LIST,

        /** An StpProperty property value */
        PROPERTY,

        /** An StpProperty.List property value */
        PROPERTY_LIST,

        /** A structure derived from Object */
        OBJECT,

        /** A list of structures derived from Object */
        OBJECT_LIST,

        /** An XML content property value. */
        XML,
        
        /** A pair of long values specifying a range. */
        RANGE;
    }

    /**
     * An enumerator denoting the type of this property's VALUE meta-property
     * value.
     */
    public static final MetaPropertyName<Type> TYPE =
        new MetaPropertyName<Type>(StpExBase.METADATA_NAMESPACE, "type");

    /**
     * Returns the value of the {@link #TYPE TYPE} meta-property as defined by
     * this StpProperty instance.
     *
     * @return  A Type enumeration representing the type of this property's
     *          VALUE meta-property.
     *
     * @throws  WvcmException  if this StpProperty instance does not define a value
     *                         for the {@link #TYPE TYPE} meta-property.
     */
    public Type getType()
        throws WvcmException;

    /** A PropertyName object for this property. */
    public static final MetaPropertyName<PropertyName<?>> PROPERTY_NAME =
        new MetaPropertyName<PropertyName<?>>(StpExBase.METADATA_NAMESPACE, "property-name");

    /**
     * Returns the value of the {@link #PROPERTY_NAME PROPERTY_NAME}
     * meta-property as defined by this StpProperty instance. Note that this
     * meta-property is always defined by a property and does not need to be
     * requested.
     *
     * @return  This object as a PropertyName.
     */
    public PropertyName<T> getPropertyName();

    /** The simple name for this property. */
    public static final MetaPropertyName<String> NAME =
        new MetaPropertyName<String>(StpExBase.METADATA_NAMESPACE, "name");

    /**
     * Returns the value of the {@link #NAME NAME} meta-property as defined by
     * this StpProperty instance. Note that this meta-property is always defined
     * by a property and does not need to be requested.
     * 
     * @return A string representing the simple name of this property's
     *         PropertyName
     */
    public String getName();

    /** The namespace for this property. */
    public static final MetaPropertyName<String> NAMESPACE =
        new MetaPropertyName<String>(StpExBase.METADATA_NAMESPACE, "namespace");

    /**
     * Returns the value of the {@link #NAMESPACE NAMESPACE} meta-property as
     * defined by this StpProperty instance. Note that this meta-property is
     * always defined by a property and does not need to be requested.
     * 
     * @return A String representing the namespace in which this property is
     *         defined.
     */
    public String getNamespace();

    /** The resource of which this is a property. */
    public static final MetaPropertyName<Resource> RESOURCE =
        new MetaPropertyName<Resource>(StpExBase.METADATA_NAMESPACE, "resource");

    /**
     * Returns the value of the {@link #RESOURCE RESOURCE} meta-property as
     * defined by this StpProperty instance.
     * 
     * @return A Resource proxy for the resource this property belongs to.
     * 
     * @throws WvcmException if this StpProperty instance does not define a
     *             value for the {@link #RESOURCE RESOURCE} meta-property
     *             component.
     */
    public Resource getResource()
        throws WvcmException;

    /**
     * A metric relating to the size of the property value, such as the number
     * of characters in a string, or the number of items in a list. Not all
     * properties define this meta-property; see the property documentation.
     */
    public static final MetaPropertyName<Long> SIZE =
        new MetaPropertyName<Long>(StpExBase.METADATA_NAMESPACE, "size");

    /**
     * Returns the value of the {@link #SIZE SIZE} meta-property as defined by
     * this StpProperty instance.
     * 
     * @return An non-negative integer.
     * 
     * @throws WvcmException if this StpProperty instance does not define a
     *             value for the {@link #SIZE SIZE} meta-property component.
     */
    public long getSize()
        throws WvcmException;

    /**
     * Whether or not the value of this property is empty/null. Not all
     * properties define this meta-property; see the property documentation.
     */
    public static final MetaPropertyName<Boolean> IS_EMPTY =
        new MetaPropertyName<Boolean>(StpExBase.METADATA_NAMESPACE, "is-empty");

    /**
     * Returns the value of the {@link #IS_EMPTY IS_EMPTY} meta-property as
     * defined by this StpProperty instance.
     * 
     * @return <b>true</b> if the value of this StpProperty is empty or null;
     *         <b> false</b> otherwise.
     * 
     * @throws WvcmException if this StpProperty instance does not define a
     *             value for the {@link #IS_EMPTY IS_EMPTY} meta-property
     *             component.
     */
    public boolean getIsEmpty()
        throws WvcmException;
}
